/*
* This confidential and proprietary software may be used only as
* authorized by licensing agreement from Multicamera.Systems LLC
* Copyright (C) 2017-2020 Multicamera.Systems LLC - All Rights Reserved
* You should have received a copy of the our license with this file.
* If not, please write to: info@multicamera.systems
*
* This code and information is provided "AS IS" without warranty of any kind,
* either expressed or implied, including but not limited to the implied
* warranties of merchantability and/or fitness for a particular purpose.
*/

#pragma once
#include <stdint.h>
#include <list>
#include "frame_desc.h"

#pragma pack(push)


typedef void(*PFN_LOG)(uint32_t LogLevel, char *szStr, ...);

typedef void(*PFN_RELEASE_STREAM_INPUT)(void *Ref, FRAME_DESC *pFrameDesc);

typedef void(*PFN_DELIVER_STREAM_OUTPUT)(void *Ref, FRAME_DESC *pFrameDesc);
typedef void(*PFN_RELEASE_STREAM_OUTPUT)(void *Ref, FRAME_DESC *pFrameDesc);



typedef struct
{
    uint32_t Zero;
    uint32_t uiGlobalID;

    INPUT_STREAM_PARAMETERS *pInputParameters;
    OUTPUT_STREAM_PARAMETERS *pOutputParameters;
    uint64_t ullTimeStampResolution;

    PFN_RELEASE_STREAM_INPUT pfnReleaseInput;
    void *ReleaseInputRef;

    PFN_DELIVER_STREAM_OUTPUT pfnDeliverOutput;
    void *DeliverOutputRef;

    // OUT
    PFN_RELEASE_STREAM_OUTPUT pfnReleaseOutput;
    void *ReleaseOutputRef;

} STREAM_CONFIGURATION;




typedef enum
{
    STREAM_PROCESS_RESULT__ERROR,
    STREAM_PROCESS_RESULT__DONE,
    STREAM_PROCESS_RESULT__PENDING

} STREAM_PROCESS_RESULT;



class Stream
{
public:
    virtual ~Stream() = 0;

    // checks if plugin is hardware based for specified configuration and does all resource management (threading, output buffers) by itself
    // can be called before Init()
    virtual bool IsHardwareBased(void *pDeviceParameters, void *pStreamParameters) = 0;

    // called once to setup a stream
    virtual bool Init(STREAM_CONFIGURATION *pStreamConfiguration) = 0;

    virtual void DeInit(void) = 0;

    // checks if parameters can be changed to specfied one
    // returns:
    // +1 if stream has to be stopped before changeing parameters
    // 0 if can be changed on the fly
    // -1 if parameters are not supported
    virtual int32_t CanUpdateParameters(void *pStreamParameters) = 0;

    // called to update streaming parameters
    virtual bool SetParameters(void *pStreamParameters) = 0;

    //  returns required output buffer size on success
    virtual uint32_t GetRequiredOutputBufferSize(void) = 0;

    // called with new data to be processed
    // if FRAME_DESC::pData == NULL then all pending async operation should be flushed
    // in async mode memory pointed by pInputFrameDesc will stay valid until it is released by STREAM_CONFIGURATION::pfnReleaseInput call
    // pOutputDataDesc can be NULL if plugin configured to self-manage output buffers
    // if error occured during async operation output frames should have FRAME_DESC::uiFrameSize == 0
    virtual STREAM_PROCESS_RESULT Process(FRAME_DESC *pInputFrameDesc, FRAME_DESC *pOutputFrameDesc) = 0;

    // waits till all pending operations finishes and all buffers are delivered and returned
    virtual void WaitForIdle(uint32_t pause_time_ms) = 0;
    
    // releases output buffer delivered to client by Process()
    virtual void ReleaseOutputBuffer(FRAME_DESC *pFrameDesc) = 0;
};



// called once when library is initialized
// Callee is responsible for freeing Caps structure
bool CreateStreamer(PFN_LOG pfnLog, void **ppCaps);

// called once when library is unloaded
void DestroyStreamer(void);

// called to create new stream
Stream * CreateStream(void);

// called to check of specified parameters are supported
bool CheckParameters(void* pDeviceParametrs, void *pStreamParameters);


#define CreateStreamerFuncName "CreateStreamer"
#define DestroyStreamerFuncName "DestroyStreamer"
#define CreateStreamFuncName "CreateStream"
#define CheckParametersFuncName "CheckParameters"


// creates new streamer object
typedef bool (*PFN_CREATE_STREAMER)(PFN_LOG pfnLog, void **ppCaps);

// destroys streamer object
typedef void (*PFN_DESTROY_STREAMER)(void);

// creates new stream
typedef Stream * (*PFN_CREATE_STREAM)(void);

// called to check of specified parameters are supported
typedef bool(*PFN_CHECK_PARAMETERS)(void* pDeviceParametrs, void *pStreamParameters);


#pragma pack(pop)