/*
* This confidential and proprietary software may be used only as
* authorized by licensing agreement from Multicamera.Systems LLC
* Copyright (C) 2017-2020 Multicamera.Systems LLC - All Rights Reserved
* You should have received a copy of the our license with this file.
* If not, please write to: info@multicamera.systems
*
* This code and information is provided "AS IS" without warranty of any kind,
* either expressed or implied, including but not limited to the implied
* warranties of merchantability and/or fitness for a particular purpose.
*/

#pragma once
#include "win_lnx.h"
#include <atomic>



template <class ObjType> class TSList;

template <class ObjType> class ObjLink
{
public:
    ObjLink *pToHead = NULL;
    ObjLink *pToTail = NULL;

    ObjType *pObj;

    class TSList<ObjType> *pOwner = NULL;
};



template <class ObjType, int MaxQueues> class ObjMLink
{
public:
    ObjMLink(ObjType *p)
    {
        for (uint32_t i = 0; i < MaxQueues; i++)
        {
            Link[i].pObj = p;
        }
    }

public:
    unsigned long long ullFreeLinkSlotMask = (1ULL << MaxQueues) - 1;

    ObjLink<ObjType> Link[MaxQueues];

    uintptr_t SideDataStorage[MaxQueues];
};




template <class ObjType> class TSList
{
public:
    TSList()
    {
        m_szName[0] = 0;
        m_pOwner = NULL;
        OS_InitializeGuardedCodeSection(&m_gsLock);
        m_uiCount.store(0);
        m_pHead = NULL;
        m_pTail = NULL;
    }

    TSList(const char *szName, void *pOwner)
    {
        strncpy(m_szName, szName, _countof(m_szName));
        m_szName[_countof(m_szName)-1] = 0;
        m_pOwner = pOwner; // used just for easier debugging
        OS_InitializeGuardedCodeSection(&m_gsLock);
        m_uiCount.store(0);
        m_pHead = NULL;
        m_pTail = NULL;
    }

    ~TSList()
    {
        clear();
        OS_DeleteGuardedCodeSection(&m_gsLock);
    }

    //
    bool empty(void)
    {
        return m_uiCount == 0;
    }

    //
    uint32_t size(void)
    {
        return m_uiCount;
    }

    //
    void clear(void)
    {
        OS_EnterGuardedCodeSection(&m_gsLock);

        while (pop_front());

        OS_LeaveGuardedCodeSection(&m_gsLock);
    }

    // returns true if queue was empty
    bool push_front(ObjType *pNewObj, uintptr_t SideData = 0)
    {
        bool res;

        OS_EnterGuardedCodeSection(&m_gsLock);

        unsigned long ulLinkSel;
#ifdef _WIN32
        if (!_BitScanForward64(&ulLinkSel, pNewObj->ullFreeLinkSlotMask))
#else
        ulLinkSel = __builtin_ffsll(pNewObj->ullFreeLinkSlotMask) - 1;
        if (ulLinkSel == (unsigned long )-1)
#endif
        {
            OS_DebugBreak(); // increase MaxQueues, or uninitialized Obj is used
            return false;
        }

        // mark it as used
        pNewObj->ullFreeLinkSlotMask &= ~(1ULL << ulLinkSel);

        ObjLink<ObjType> *pNewLink = &pNewObj->Link[ulLinkSel];

        if (m_pHead)
        {
            m_pHead->pToHead = pNewLink;
        }
        else
        {
            m_pTail = pNewLink;
        }

        pNewLink->pToTail = m_pHead;

        m_pHead = pNewLink;

        pNewLink->pToHead = NULL;
        pNewLink->pOwner = this;

        pNewObj->SideDataStorage[ulLinkSel] = SideData;

        res = (m_uiCount == 0);
        m_uiCount++;

        OS_LeaveGuardedCodeSection(&m_gsLock);

        return res;
    }

    // returns true if queue was empty
    bool push_back(ObjType *pNewObj, uintptr_t SideData = 0)
    {
        bool res;

        OS_EnterGuardedCodeSection(&m_gsLock);

        unsigned long ulLinkSel;
#ifdef _WIN32
        if (!_BitScanForward64(&ulLinkSel, pNewObj->ullFreeLinkSlotMask))
#else
        ulLinkSel = __builtin_ffsll(pNewObj->ullFreeLinkSlotMask) - 1;
        if (ulLinkSel == (unsigned long)-1)
#endif
        {
            OS_DebugBreak(); // increase MaxQueues, or uninitialized Obj is used
            return false;
        }

        // mark it as used
        pNewObj->ullFreeLinkSlotMask &= ~(1ULL << ulLinkSel);

        ObjLink<ObjType> *pNewLink = &pNewObj->Link[ulLinkSel];

        if (m_pTail)
        {
            m_pTail->pToTail = pNewLink;
        }
        else
        {
            m_pHead = pNewLink;
        }

        pNewLink->pToHead = m_pTail;

        m_pTail = pNewLink;

        pNewLink->pToTail = NULL;
        pNewLink->pOwner = this;

        pNewObj->SideDataStorage[ulLinkSel] = SideData;

        res = (m_uiCount == 0);
        m_uiCount++;

        OS_LeaveGuardedCodeSection(&m_gsLock);

        return res;
    }

    //
    ObjType * pop_front(uintptr_t *pullSideData = NULL)
    {
        OS_EnterGuardedCodeSection(&m_gsLock);

        ObjType *pObj = NULL;

        if (m_pHead)
        {
            ObjLink<ObjType> *p = m_pHead;

            m_pHead = p->pToTail;
            if (m_pHead)
            {
                m_pHead->pToHead = NULL;
            }
            else
            {
                m_pTail = NULL;
            }

            p->pToHead = NULL;
            p->pToTail = NULL;
            p->pOwner = NULL;

            pObj = p->pObj;

            unsigned long ulLinkSel = (unsigned long)(p - &pObj->Link[0]);

            if (pullSideData)
            {
                *pullSideData = pObj->SideDataStorage[ulLinkSel];
            }

            pObj->ullFreeLinkSlotMask |= (1ULL << ulLinkSel);
            m_uiCount--;
        }

        OS_LeaveGuardedCodeSection(&m_gsLock);

        return pObj;
    }

    //
    ObjType * pop_back(uintptr_t *pullSideData = NULL)
    {
        OS_EnterGuardedCodeSection(&m_gsLock);

        ObjType *pObj = NULL;

        if (m_pTail)
        {
            ObjLink<ObjType> *p = m_pTail;

            m_pTail = p->pToHead;
            if (m_pTail)
            {
                m_pTail->pToTail = NULL;
            }
            else
            {
                m_pHead = NULL;
            }

            p->pToHead = NULL;
            p->pToTail = NULL;
            p->pOwner = NULL;

            pObj = p->pObj;

            unsigned long ulLinkSel = (unsigned long)(p - &pObj->Link[0]);

            if (pullSideData)
            {
                *pullSideData = pObj->SideDataStorage[ulLinkSel];
            }

            pObj->ullFreeLinkSlotMask |= (1ULL << ulLinkSel);
            m_uiCount--;
        }

        OS_LeaveGuardedCodeSection(&m_gsLock);

        return pObj;
    }


    //
    // returns true if queue was empty
    template<class _Pr2>
    bool insert_before_conditional(_Pr2 _Pred, ObjType *pNewObj, uintptr_t ullSideData = 0)
    {
        bool res;

        OS_EnterGuardedCodeSection(&this->m_gsLock);

        unsigned long ulLinkSel;
#ifdef _WIN32
        if (!_BitScanForward64(&ulLinkSel, pNewObj->ullFreeLinkSlotMask))
#else
        ulLinkSel = __builtin_ffsll(pNewObj->ullFreeLinkSlotMask) - 1;
        if (ulLinkSel == (unsigned long)-1)
#endif
        {
            OS_DebugBreak(); // increase MaxQueues, or uninitialized Obj is used
            return false;
        }

        // mark it as used
        pNewObj->ullFreeLinkSlotMask &= ~(1ULL << ulLinkSel);

        ObjLink<ObjType> *pNewLink = &pNewObj->Link[ulLinkSel];

        pNewLink->pObj->SideDataStorage[ulLinkSel] = ullSideData;

        if (!m_pHead)
        {
            m_pHead = pNewLink;
            m_pTail = pNewLink;
            pNewLink->pToHead = NULL;
            pNewLink->pToTail = NULL;
            res = true;
        }
        else
        {
            res = false;
            ObjLink<ObjType> *pInsertBefore = m_pHead;

            while (pInsertBefore)
            {
                if (_Pred(pInsertBefore->pObj, pNewObj))
                {
                    break;
                }

                pInsertBefore = pInsertBefore->pToTail;
            }

            if (pInsertBefore)
            {
                pNewLink->pToTail = pInsertBefore;
                pNewLink->pToHead = pInsertBefore->pToHead;
                pInsertBefore->pToHead = pNewLink;
                if (m_pHead == pInsertBefore)
                {
                    m_pHead = pNewLink;
                }
                else
                {
                    pNewLink->pToHead->pToTail = pNewLink;
                }
            }
            else
            {
                m_pTail->pToTail = pNewLink;
                pNewLink->pToHead = this->m_pTail;
                pNewLink->pToTail = NULL;
                m_pTail = pNewLink;
            }
        }

        pNewLink->pOwner = this;

        m_uiCount++;

        OS_LeaveGuardedCodeSection(&m_gsLock);

        return res;
    }

    //
    template<class _Pr2>
    ObjType * pop_conditional(void *pComp, _Pr2 _Pred, uint32_t max_check_depth = UINT32_MAX, uintptr_t *pullSideData = NULL)
    {
        OS_EnterGuardedCodeSection(&this->m_gsLock);

        ObjType *pObj = NULL;

        uint32_t depth = 0;

        ObjLink<ObjType> *p = this->m_pHead;
        while (p)
        {
            if (_Pred(p->pObj, pComp))
            {
                if (p->pToTail)
                {
                    p->pToTail->pToHead = p->pToHead;
                }
                else
                {
                    m_pTail = p->pToHead;
                }

                if (p->pToHead)
                {
                    p->pToHead->pToTail = p->pToTail;
                }
                else
                {
                    m_pHead = p->pToTail;
                }

                p->pToHead = NULL;
                p->pToTail = NULL;
                p->pOwner = NULL;

                pObj = p->pObj;

                unsigned long ulLinkSel = (unsigned long)(p - &pObj->Link[0]);
                if (pullSideData)
                {
                    *pullSideData = pObj->SideDataStorage[ulLinkSel];
                }

                pObj->ullFreeLinkSlotMask |= (1ULL << ulLinkSel);
                m_uiCount--;

                break;
            }
            else
            if (++depth == max_check_depth)
            {
                break;
            }

            p = p->pToTail;
        }

        OS_LeaveGuardedCodeSection(&m_gsLock);

        return pObj;
    }

    
    // returns true if queue was empty
    bool move_all_to_other_back(TSList *pOtherList)
    {
        bool res = true;

        OS_EnterGuardedCodeSection(&m_gsLock);

        if (m_pHead)
        {
            res = pOtherList->move_to_back(m_pHead, m_pTail, m_uiCount);
            m_pHead = NULL;
            m_pTail = NULL;
            m_uiCount = 0;
        }

        OS_LeaveGuardedCodeSection(&m_gsLock);

        return res;
    }

    // returns true if queue we are moving items to was empty
    bool move_all_from_other(TSList *pOtherList)
    {
        return pOtherList->move_all_to_other_back(this);
    }

protected:
    bool move_to_back(ObjLink<ObjType> *pOtherHead, ObjLink<ObjType> *pOtherTail, uint32_t uiOtherCount)
    {
        bool res;

        OS_EnterGuardedCodeSection(&m_gsLock);

        ObjLink<ObjType> *p = pOtherHead;
        while (p)
        {
            p->pOwner = this;
            p = p->pToTail;
        }

        if (m_pHead)
        {
            m_pTail->pToTail = pOtherHead;
            pOtherHead->pToHead = m_pTail;
        }
        else
        {
            m_pHead = pOtherHead;
        }

        m_pTail = pOtherTail;

        res = (m_uiCount == 0);

        m_uiCount += uiOtherCount;

        OS_LeaveGuardedCodeSection(&m_gsLock);

        return res;
    }


protected:
    ObjLink<ObjType> *m_pHead = NULL;
    ObjLink<ObjType> *m_pTail = NULL;

    std::atomic<uint32_t> m_uiCount;

    OS_GUARDED_CODE_SECTION m_gsLock;

    char m_szName[64];
    void *m_pOwner;
};

