/*
* This confidential and proprietary software may be used only as
* authorized by licensing agreement from Multicamera.Systems LLC
* Copyright (C) 2017-2020 Multicamera.Systems LLC - All Rights Reserved
* You should have received a copy of the our license with this file.
* If not, please write to: info@multicamera.systems
*
* This code and information is provided "AS IS" without warranty of any kind,
* either expressed or implied, including but not limited to the implied
* warranties of merchantability and/or fitness for a particular purpose.
*/

//
// This sample demonstrates the usage of GPUSqueeze library
// It generates two same resolution elementary (without MP4/MKV container) HEVC video streams from a static bayer images
//
// Attention! If you use consumer grade nVidia GPUs (GTX/RTX) you have the following limitation imposed by nVidia driver:
//            you cannot have several streams with different input/output resolution and/or encoding parameters,
//            but you can have multiple identical streams. 
//

#define _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include "../include/GPUSqueezeLib.h"
#include <vector>
#include <mutex>
#include <conio.h>
#include <varargs.h>


std::mutex g_LogMutex;


// this function is a callback
// it is called by GPUSqueeze library for logging (used log levels [0..2], always/info/error)
void Log(uint32_t LogLevel, char *szStr, ...)
{
    g_LogMutex.lock();
    va_list args;
    va_start(args, szStr);
    _vcprintf(szStr, args);
    printf("\n");
    va_end(args);
    g_LogMutex.unlock();
}



// this structure defines input video streams, in this sample we use just one input
INPUT_STREAM_PARAMETERS g_InputStreamParameters[] =
{
    {
        {PIXEL_FORMAT__BAYER_RG, 8, 8},     // input image format
        {COMPRESSION_TYPE__UNCOMPRESSED},   // input compression type
        1920, 1080, 30.f                    // input resulution and frame rate
    },
    {
        {PIXEL_FORMAT__BAYER_RG, 8, 8},     // input image format
        {COMPRESSION_TYPE__UNCOMPRESSED},   // input compression type
        1920, 1080, 30.f                    // input resulution and frame rate
    }
};


// this structure defines output video streams, we use  identical streams
// if you have Quadro board (or special driver which allows to do it) you may have any number of different streams
OUTPUT_STREAM_PARAMETERS g_OutputStreamParameters[] =
{
    {
        1920, 1080,                             // output resulution
        { 0, 0, 0, 0 },                         // input window [x, y, width, height], zeros default to full input resolution
        { {COMPRESSION_TYPE__H265, 0, 1, 8},    // output compression type [type (AVC or HEVC), GOP size (0==auto), chroma samples per luma (valid values 1(420) and 4(444)), bit precision)
        COMPRESSION_FLAG__QUALITY_ID_VALID, COMPRESSION_QUALITY_ID__MEDIUM, 0.f, 0 }, // compression quality settings [flags, quality preset, N/A, bit rate]
    },
    {
        1920, 1080,                             // output resulution
        { 0, 0, 0, 0 },                         // input window [x, y, width, height], zeros default to full input resolution
        { {COMPRESSION_TYPE__H265, 0, 1, 8},    // output compression type [type (AVC or HEVC), GOP size (0==auto), chroma samples per luma (valid values 1(420) and 4(444)), bit precision)
        COMPRESSION_FLAG__QUALITY_ID_VALID, COMPRESSION_QUALITY_ID__MEDIUM, 0.f, 0 }, // compression quality settings [flags, quality preset, N/A, bit rate]
    }
};


// each object represents one video stream
GPUVideoStream *g_pGPUStream[_countof(g_OutputStreamParameters)];


// files and misc
void *g_pInputImage[2] = { NULL };
uint32_t g_uiInputImageSize = 1920 * 1080;  // fixed in this sample
FILE *g_hOutFile[_countof(g_OutputStreamParameters)] = { 0 };
uint32_t g_uiStreamCount = _countof(g_OutputStreamParameters);
uint32_t g_uiFrameCount = 1000;


// input buffer descriptor list
// all operation with FRAME_DESC_QUEUE are atomic
FRAME_DESC_QUEUE g_FrameDescList;

// null frame descriptor used for flushing pipeline
FRAME_DESC g_NullFrame;


// here we store address of the function retured by GPUSqueeze library after initialization
// this function should be called to release buffer with compressed image back to the library after we done writing it
PFN_RELEASE_STREAM_OUTPUT g_pReleaseOutput[_countof(g_OutputStreamParameters)];

// here we store the reference we need to pass to the g_pReleaseOutput[]() function
// there reference is also supplied to us after initialization step
void* g_ReleaseOutputRef[_countof(g_OutputStreamParameters)];



// this function called by GPUSqueeze library for each output compressed frame
void DeliverOutput(void *Ref, FRAME_DESC *pFrameDesc)
{
    // this is our reference we passed to the library during initialization
    // we use stream_id as a reference
    uint64_t stream_id = (uint64_t)Ref;

    // writeout data
    fwrite(pFrameDesc->pData, 1, pFrameDesc->uiFrameSize, g_hOutFile[stream_id]);

    // release buffer back to the library
    void* LibRef = g_ReleaseOutputRef[stream_id];
    g_pReleaseOutput[stream_id](LibRef, pFrameDesc);
}



DWORD __stdcall SubmitThreadProc(void* lpThreadParameter)
{
    uint64_t stream_id = (uint64_t)lpThreadParameter;

    bool bErr = false;

    for (uint32_t frame_number = 0; (frame_number < g_uiFrameCount) && !bErr; frame_number++)
    {
        FRAME_DESC *pFrameDesc;

        // get free input frame descriptor
        do
        {
            pFrameDesc = g_FrameDescList.pop_front();
            if (!pFrameDesc)
            {
                Sleep(0);
            }
        } while (!pFrameDesc);

        // copy our static image
        memcpy(pFrameDesc->pData, g_pInputImage[stream_id], g_uiInputImageSize);
        pFrameDesc->uiFrameSize = g_uiInputImageSize;

        // it is required that frame numbers are updated for each new frame
        pFrameDesc->ullFrameNumber = frame_number;

        // set time stamp to something - it will be passed through to the DeliverOutput() callback
        pFrameDesc->ullTimeStamp = (uint64_t)(frame_number * (1000000 / g_InputStreamParameters[0].fFrameRate));

        // call the library to process the frame (processing is asynchronous)
        if (g_pGPUStream[stream_id]->Process(pFrameDesc) == STREAM_PROCESS_RESULT__ERROR)
        {
            bErr = true;
        }
    }

    // flush pipeline (the WaitForIdle() call also flushes the pipeline, but we put this here to show how to flush it manually)
    g_pGPUStream[stream_id]->Process(&g_NullFrame);

    // and wait till done
    g_pGPUStream[stream_id]->WaitForIdle();
    

    return 0;
}



void ReleaseInput(void *Ref, FRAME_DESC *pFrameDesc)
{
    uint64_t stream_id = (uint64_t)Ref;

    g_FrameDescList.push_back(pFrameDesc);
}




int main()
{
    void *pKeyData = NULL;
    uint32_t key_size = 0;

#if 0
    // read certificate file to remove watermark
    {
        FILE *f = fopen("C:\\ProgramData\\MCS\\Cert\\mcs.cert.g", "rb");
        fseek(f, 0, SEEK_END);
        key_size = ftell(f);
        fseek(f, 0, SEEK_SET);
        pKeyData = malloc(key_size);
        fread(pKeyData, key_size, 1, f);
        fclose(f);
    }
#endif

    // load library
    HMODULE hGPULib = LoadLibrary("../bin/GPUSqueeze");
    if (!hGPULib)
    {
        Log(0, "No GPUSqueeze library found (%u)", GetLastError());
        return 0;
    }

    PFN_CREATE_STREAMER_GPU pfnCreateStreamerGPU = (PFN_CREATE_STREAMER_GPU)GetProcAddress(hGPULib, CreateStreamerGPUFuncName);
    PFN_DESTROY_STREAMER_GPU pfnDestroyStreamerGPU = (PFN_DESTROY_STREAMER_GPU)GetProcAddress(hGPULib, DestroyStreamerGPUFuncName);
    PFN_CREATE_STREAM_GPU pfnCreateStreamGPU = (PFN_CREATE_STREAM_GPU)GetProcAddress(hGPULib, CreateStreamGPUFuncName);
    PFN_CHECK_PARAMETERS_GPU pfnCheckParametersGPU = (PFN_CHECK_PARAMETERS_GPU)GetProcAddress(hGPULib, CheckParametersGPUFuncName);

    // open the library
    if (!pfnCreateStreamerGPU(Log, 0, pKeyData, key_size))
    {
        Log(0, "Failed to init GPUSqueeze library");
        return 0;
    }

    // check that input parameters are supported
    for (uint32_t stream_id = 0; stream_id < g_uiStreamCount; stream_id++)
    {
        if (!pfnCheckParametersGPU(&g_InputStreamParameters[0], &g_OutputStreamParameters[stream_id]))
        {
            Log(0, "Unsupported parameters for stream %u", stream_id);
            return 0;
        }
    }

    // read the input image
    g_pInputImage[0] = malloc(g_uiInputImageSize);
    FILE *f = fopen("testimageHDBayerRG.raw", "rb");
    fread(g_pInputImage[0], 1, g_uiInputImageSize, f);
    fclose(f);

    // make a second image, invert it to make it different
    g_pInputImage[1] = malloc(g_uiInputImageSize);
    for (uint32_t i = 0; i < g_uiInputImageSize; i++) ((uint8_t*)(g_pInputImage[1]))[i] = ~((uint8_t*)(g_pInputImage[0]))[i];

    // open output files
    for (uint32_t stream_id = 0; stream_id < g_uiStreamCount; stream_id++)
    {
        char szName[32];
        sprintf(szName, "%u.h265", stream_id);
        g_hOutFile[stream_id] = fopen(szName, "wb");
    }

    uint32_t uiInputBufferCount = 0;

    for (uint32_t stream_id = 0; stream_id < g_uiStreamCount; stream_id++)
    {
        bool bOk = false;

        g_pGPUStream[stream_id] = pfnCreateStreamGPU();
        if (g_pGPUStream[stream_id])
        {
            g_OutputStreamParameters[stream_id].CodecInfo.h26x.bFixupNALUSize = false;  // set it to true if you do transcoding and you compressed video input
                                                                                        // has NALU size instead of NALU marker

            g_OutputStreamParameters[stream_id].CodecInfo.h26x.bGenerateSPSPPS = true;  // ask to embed video stream parameters at the beggining of the stream

            STREAM_CONFIGURATION StreamConfig;
            StreamConfig.Zero = 0;
            StreamConfig.uiGlobalID = stream_id + 1;    // used just for logging
            StreamConfig.pInputParameters = &g_InputStreamParameters[stream_id];    // pointer to input stream description
            StreamConfig.pOutputParameters = &g_OutputStreamParameters[stream_id];  // pointer to output stream description
            StreamConfig.ullTimeStampResolution = 0; // not used
            StreamConfig.pfnReleaseInput = ReleaseInput;                    // our function to be called by the GPUSqueeze library to release input buffer
            StreamConfig.ReleaseInputRef = (void*)(uintptr_t)stream_id;     // our referent for above callback
            StreamConfig.pfnDeliverOutput = DeliverOutput;                  // our function to be called by the GPUSqueeze library to deliver output buffer
            StreamConfig.DeliverOutputRef = (void*)(uintptr_t)stream_id;    // our referent for above callback

            bOk = g_pGPUStream[stream_id]->Init(&StreamConfig);

            // save returned callback address and reference for output buffer release
            g_pReleaseOutput[stream_id] = StreamConfig.pfnReleaseOutput;
            g_ReleaseOutputRef[stream_id] = StreamConfig.ReleaseOutputRef;

            // the following structure contains VPS/SPS/PPS info you may need to use for generating video container
            // g_OutputStreamParameters[stream_id].CodecInfo.h26x

            // we need to have at least this number of input buffer
            uiInputBufferCount = max(uiInputBufferCount, g_OutputStreamParameters[stream_id].uiPipelineDepth);
        }

        if (!bOk)
        {
            Log(0, "Failed to init stream %u", stream_id);
            pfnDestroyStreamerGPU();
            return 0;
        }
    }

    // allocate more buffers than minimally required
    uiInputBufferCount *= 2;

    // allocate input buffer descriptors
    FRAME_DESC *pFrameDescArray = { 0 };
    pFrameDescArray = new FRAME_DESC[uiInputBufferCount];   

    // allocate input buffers and previously allocated set descriptors
    for (uint32_t i = 0; i < uiInputBufferCount; i++)
    {
        pFrameDescArray[i].uiFrameSize = g_uiInputImageSize;

        // allocate the buffer
        pFrameDescArray[i].pData = _aligned_malloc(pFrameDescArray[i].uiFrameSize, 4096);

        // pFrameDescArray[i].DataRef[] are placeholders for additional references if needed

        // add the description to out list
        g_FrameDescList.push_back(&pFrameDescArray[i]);
    }


    // create threads (processing will start)
    // in this sample we use separate threads to submit input frames, but you can do it in the main thread

    HANDLE hThread[_countof(g_OutputStreamParameters)];

    for (uint64_t stream_id = 0; stream_id < g_uiStreamCount; stream_id++)
    {
        hThread[stream_id] = CreateThread(NULL, 0, SubmitThreadProc, (void*)stream_id, 0, 0);
    }

    // wait until everything is finished
    for (uint64_t stream_id = 0; stream_id < g_uiStreamCount; stream_id++)
    {
        WaitForSingleObject(hThread[stream_id], INFINITE);
        CloseHandle(hThread[stream_id]);
    }


    // clean up

    for (uint64_t stream_id = 0; stream_id < g_uiStreamCount; stream_id++)
    {
        g_pGPUStream[stream_id]->DeInit();
        if (g_pGPUStream[stream_id])
        {
            delete g_pGPUStream[stream_id];
        }
    }

    pfnDestroyStreamerGPU();

    g_FrameDescList.clear();

    for (uint32_t i = 0; i < uiInputBufferCount; i++)
    {
        if (pFrameDescArray[i].pData)
        {
            _aligned_free(pFrameDescArray[i].pData);
        }
    }

    delete[] pFrameDescArray;

    for (uint32_t i = 0; i < _countof(g_pInputImage); i++)
    {
        free(g_pInputImage[i]);
    }

    for (uint32_t i = 0; i < _countof(g_hOutFile); i++)
    {
        if (g_hOutFile[i])
        {
            fclose(g_hOutFile[i]);
        }
    }

    FreeLibrary(hGPULib);

    return 1;
}


